import tkinter
import random

from srcgen import lock_controller, water_level_simulator
from src.yakindu.rx import Observer
from src import simulator

WATER_COLOR = '#1c8ce8'
SKY_COLOR = '#ffc4d4'
DOOR_COLOR = '#633f09'

TRAFFIC_LIGHT_OFF_COLOR = '#362d2d'
TRAFFIC_LIGHT_RED_COLOR = '#ff0000'
TRAFFIC_LIGHT_GREEN_COLOR = '#4be81c'

DEFAULT_WIDGET_COLOR = '#d9d9d9'

class LockDoorView:
    def __init__(self, canvas, x, scale=1):
        self.canvas = canvas
        self.door_id = canvas.create_rectangle((x-10)*scale, 70*scale, (x+10)*scale, 250*scale)
        self.flow_id = canvas.create_rectangle((x-12)*scale, 220*scale, (x+12)*scale, 240*scale, outline='')

        canvas.create_oval((x-15)*scale, 10, (x+15)*scale, 54, fill='black')
        self.red_light_id = canvas.create_oval((x-10)*scale, 12, (x+10)*scale, 32, outline='')
        self.green_light_id = canvas.create_oval((x-10)*scale, 32, (x+10)*scale, 52, outline='')

        self.set_doors(open=False)
        self.set_flow(open=False)
        self.set_red_light()

    def set_doors(self, open):
        if open:
            self.canvas.itemconfig(self.door_id, fill='', outline='black', dash=(4,6))
        else:
            self.canvas.itemconfig(self.door_id, fill=DOOR_COLOR, outline='', dash=None)

    def set_flow(self, open):
        if open:
            self.canvas.itemconfig(self.flow_id, fill=WATER_COLOR)
        else:
            self.canvas.itemconfig(self.flow_id, fill='')

    def set_green_light(self):
        self.canvas.itemconfig(self.red_light_id, fill=TRAFFIC_LIGHT_OFF_COLOR)
        self.canvas.itemconfig(self.green_light_id, fill=TRAFFIC_LIGHT_GREEN_COLOR)

    def set_red_light(self):
        self.canvas.itemconfig(self.red_light_id, fill=TRAFFIC_LIGHT_RED_COLOR)
        self.canvas.itemconfig(self.green_light_id, fill=TRAFFIC_LIGHT_OFF_COLOR)

class LockView:
    def __init__(self, parent, scale=1):
        self.scale = scale
        self.canvas = tkinter.Canvas(toplevel, bg=SKY_COLOR, width=600*scale, height=250*scale)
        # LOW side:
        self.canvas.create_rectangle(0, 200*scale, 200*scale, 250*scale, fill=WATER_COLOR, outline='')
        # MIDDLE side:
        self.middle_rectangle_id = self.canvas.create_rectangle(200*scale, 200*scale, 400*scale, 250*scale, fill=WATER_COLOR, outline='')
        # HIGH side:
        self.canvas.create_rectangle(400*scale, 100*scale, 600*scale, 250*scale, fill=WATER_COLOR, outline='')

        self.ldoor = LockDoorView(self.canvas, 200, scale)
        self.hdoor = LockDoorView(self.canvas, 400, scale)

    def set_water_lvl(self, water_lvl):
        self.canvas.coords(self.middle_rectangle_id, 200*self.scale, (250-water_lvl/10)*self.scale, 400*self.scale, 250*self.scale)


if __name__ == "__main__":
    random.seed(1)

    toplevel = tkinter.Tk()
    toplevel.resizable(0,0)

    sc = lock_controller.LockController()
    wlvlsc = water_level_simulator.WaterLevelSimulator() # our environment is also modeled as a statechart :)
    controller = simulator.Controller()
    sc.timer_service = simulator.TimerService(controller)
    wlvlsc.timer_service = simulator.TimerService(controller)

    def update_callback():
        string_simtime.set('{:.3f}'.format((controller.get_simtime_seconds())))
    sim = simulator.RealTimeSimulation(controller, toplevel, update_callback)

    toplevel.title("Lock Simulator")

    lock_view = LockView(toplevel)
    lock_view.canvas.pack(side=tkinter.TOP)

    string_simtime = tkinter.StringVar()
    string_ldoors = tkinter.StringVar(value="CLOSED")
    string_hdoors = tkinter.StringVar(value="CLOSED")
    string_lflow = tkinter.StringVar(value="CLOSED")
    string_hflow = tkinter.StringVar(value="CLOSED")
    string_lsignal = tkinter.StringVar(value="RED")
    string_hsignal = tkinter.StringVar(value="RED")
    string_sensor = tkinter.StringVar(value="")
    string_real_lvl = tkinter.StringVar(value="")
    string_sensor_status = tkinter.StringVar(value="NO FAILURE DETECTED")

    def set_doors(side, open):
        strval = "OPEN" if open else "CLOSED"
        if side == sc.LOW:
            string_ldoors.set(strval)
            lock_view.ldoor.set_doors(open)
        elif side == sc.HIGH:
            string_hdoors.set(strval)
            lock_view.hdoor.set_doors(open)

    def set_flow(side, open):
        strval = "OPEN" if open else "CLOSED"
        eventname = "open_flow" if open else "close_flow"
        sim.add_input(wlvlsc, eventname, value=side)
        if side == sc.LOW:
            string_lflow.set(strval)
            lock_view.ldoor.set_flow(open)
        elif side == sc.HIGH:
            string_hflow.set(strval)
            lock_view.hdoor.set_flow(open)

    def set_green_light(side):
        if side == sc.LOW:
            string_lsignal.set("GREEN")
            lock_view.ldoor.set_green_light()
        else:
            string_hsignal.set("GREEN")
            lock_view.hdoor.set_green_light()

    def set_red_light(side):
        if side == sc.LOW:
            string_lsignal.set("RED")
            lock_view.ldoor.set_red_light()
        else:
            string_hsignal.set("RED")
            lock_view.hdoor.set_red_light()


    def break_sensor():
        sim.add_input(wlvlsc, "toggle_sensor_broken")
        entry_sensor.config(readonlybackground='red')
        button_broken_sensor.config(state=tkinter.DISABLED)
        button_unbroken_sensor.config(state=tkinter.NORMAL)

    def unbreak_sensor():
        sim.add_input(wlvlsc, "toggle_sensor_broken")
        entry_sensor.config(readonlybackground=DEFAULT_WIDGET_COLOR) # default color
        button_broken_sensor.config(state=tkinter.NORMAL)
        button_unbroken_sensor.config(state=tkinter.DISABLED)

    sim_frame = tkinter.LabelFrame(toplevel, text="Environment")
    tkinter.Label(sim_frame, text="Real Water Level").grid(column=0, row=0)
    tkinter.Entry(sim_frame, state='readonly', width=8, textvariable=string_real_lvl, justify=tkinter.RIGHT).grid(column=1, row=0)
    tkinter.Label(sim_frame, text="cm").grid(column=2, row=0)

    tkinter.Label(sim_frame, text="Water Level Sensor").grid(column=0, row=1)
    entry_sensor = tkinter.Entry(sim_frame, state='readonly', width=8, textvariable=string_sensor, justify=tkinter.RIGHT)
    entry_sensor.grid(row=1, column=1)
    tkinter.Label(sim_frame, text="cm").grid(row=1, column=2)

    button_broken_sensor = tkinter.Button(sim_frame, text="Break Sensor",
        command=break_sensor, width=14)
    button_unbroken_sensor = tkinter.Button(sim_frame, text="Un-break Sensor",
        command=unbreak_sensor, width=14, state=tkinter.DISABLED)
    button_broken_sensor.grid(column=0, row=3, columnspan=3)
    button_unbroken_sensor.grid(column=0, row=4, columnspan=3)

    tkinter.Label(sim_frame, text="Simulated Time").grid(column=0, row=5)
    tkinter.Entry(sim_frame, state='readonly', width=8, textvariable=string_simtime, justify=tkinter.RIGHT).grid(row=5, column=1)
    tkinter.Label(sim_frame, text="s").grid(row=5, column=2)
    sim_frame.pack(side=tkinter.LEFT)

    def resume():
        sim.add_input(sc, "resume")
        button_resume.config(state=tkinter.DISABLED, bg=DEFAULT_WIDGET_COLOR)
        string_sensor_status.set("NO FAILURE DETECTED")

    request_frame = tkinter.LabelFrame(toplevel, text="Actions")
    button_change_lvl = tkinter.Button(request_frame, command=lambda: sim.add_input(sc, "request_lvl_change"), width=18)
    button_change_lvl.pack()
    button_resume = tkinter.Button(request_frame, text="Resume (make sure sensor is repaired first!)",
        command=resume, wraplength=160, width=18, state=tkinter.DISABLED)
    button_resume.pack()
    request_frame.pack(side=tkinter.LEFT)

    def set_indicator(value):
        if value:
            button_change_lvl.config(state=tkinter.DISABLED, text="Change requested")
        else:
            button_change_lvl.config(state=tkinter.NORMAL, text="Change water level")
    set_indicator(value=False)

    status_frame = tkinter.LabelFrame(toplevel, text="Status")
    tkinter.Label(status_frame, text="Low").grid(row=0, column=1)
    tkinter.Label(status_frame, text="High").grid(row=0, column=2)
    tkinter.Label(status_frame, text="Doors").grid(row=1, column=0)
    tkinter.Label(status_frame, text="Flow").grid(row=2, column=0)
    tkinter.Label(status_frame, text="Signal").grid(row=3, column=0)
    tkinter.Entry(status_frame, state='readonly', width=8, textvariable=string_ldoors).grid(row=1, column=1)
    tkinter.Entry(status_frame, state='readonly', width=8, textvariable=string_hdoors).grid(row=1, column=2)
    tkinter.Entry(status_frame, state='readonly', width=8, textvariable=string_lflow).grid(row=2, column=1)
    tkinter.Entry(status_frame, state='readonly', width=8, textvariable=string_hflow).grid(row=2, column=2)
    tkinter.Entry(status_frame, state='readonly', width=8, textvariable=string_lsignal).grid(row=3, column=1)
    tkinter.Entry(status_frame, state='readonly', width=8, textvariable=string_hsignal).grid(row=3, column=2)
    tkinter.Label(status_frame, text="Sensor").grid(row=4, column=0)
    tkinter.Entry(status_frame, state='readonly', width=18, textvariable=string_sensor_status).grid(row=4, column=1, columnspan=2)
    status_frame.pack(side=tkinter.LEFT)


    ### Output event handlers ###

    class CallbackObserver(Observer):
        def __init__(self, event, callback):
            self.event = event
            self.callback = callback

        def next(self, value=None):
            if value is None:
                self.callback()
            else:
                self.callback(value)

    def on_water_level_reading(water_level):
        noisy_water_level = water_level + int(random.random()*10)
        sim.add_input(sc, "water_lvl", value=noisy_water_level)
        string_sensor.set(noisy_water_level)

    def on_real_water_level(water_level):
        string_real_lvl.set(water_level)
        lock_view.set_water_lvl(water_level)

    def set_sensor_broken():
        button_resume.config(state=tkinter.NORMAL, bg='yellow')
        string_sensor_status.set("FAILURE DETECTED")

    sc.set_request_pending_observable.subscribe(CallbackObserver("setIndicator", set_indicator))
    sc.open_flow_observable.subscribe(CallbackObserver("openFlow", lambda side: set_flow(side, True)))
    sc.close_flow_observable.subscribe(CallbackObserver("closeFlow", lambda side: set_flow(side, False)))
    sc.open_doors_observable.subscribe(CallbackObserver("openDoors", lambda side: set_doors(side, True)))
    sc.close_doors_observable.subscribe(CallbackObserver("closeDoors", lambda side: set_doors(side, False)))
    sc.green_light_observable.subscribe(CallbackObserver("greenLight", set_green_light))
    sc.red_light_observable.subscribe(CallbackObserver("redLight", set_red_light))
    sc.set_sensor_broken_observable.subscribe(CallbackObserver("setSensorBroken", set_sensor_broken))
    wlvlsc.sensor_reading_observable.subscribe(CallbackObserver(None, on_water_level_reading))
    wlvlsc.real_water_level_observable.subscribe(CallbackObserver(None, on_real_water_level))


    controller.start()

    # Enter default states
    sc.enter()
    wlvlsc.enter()

    sim.interrupt() # schedule first wakeup
    toplevel.mainloop() # everything is controlled by tkinter's eventloop
